use crate::value::{Value, NestedValue, Map, Domain};
use serde::Serialize;
use crate::value::error::{Error, Result};
use std::marker::PhantomData;

pub struct Serializer<N: NestedValue<D>, D: Domain> {
    phantom: PhantomData<(N, D)>,
}

pub struct SerializeDictionary<N: NestedValue<D>, D: Domain> {
    next_key: Option<N>,
    items: Map<N,N>,
    phantom: PhantomData<D>,
}

pub struct SerializeRecord<N: NestedValue<D>, D: Domain> {
    name: &'static str,
    vec: Vec<N>,
    phantom: PhantomData<D>,
}

pub struct SerializeSequence<N: NestedValue<D>, D: Domain> {
    vec: Vec<N>,
    phantom: PhantomData<D>,
}

impl<N: NestedValue<D>, D: Domain> serde::Serializer for Serializer<N, D> {
    type Ok = N;
    type Error = Error<D>;
    type SerializeSeq = SerializeSequence<N, D>;
    type SerializeTuple = SerializeRecord<N, D>;
    type SerializeTupleStruct = SerializeRecord<N, D>;
    type SerializeTupleVariant = SerializeRecord<N, D>;
    type SerializeMap = SerializeDictionary<N, D>;
    type SerializeStruct = SerializeRecord<N, D>;
    type SerializeStructVariant = SerializeRecord<N, D>;

    fn serialize_bool(self, v: bool) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_i8(self, v: i8) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_i16(self, v: i16) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_i32(self, v: i32) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_i64(self, v: i64) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_u8(self, v: u8) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_u16(self, v: u16) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_u32(self, v: u32) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_u64(self, v: u64) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_f32(self, v: f32) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_f64(self, v: f64) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_char(self, v: char) -> Result<Self::Ok, D> {
        Ok(Value::simple_record("UnicodeScalar", vec![Value::from(v as u32).wrap()]).wrap())
    }

    fn serialize_str(self, v: &str) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_bytes(self, v: &[u8]) -> Result<Self::Ok, D> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_none(self) -> Result<Self::Ok, D> {
        Ok(Value::simple_record("None", vec![]).wrap())
    }

    fn serialize_some<T: ?Sized>(self, v: &T) -> Result<Self::Ok, D> where T: Serialize {
        Ok(Value::simple_record("Some", vec![to_value(v)?]).wrap())
    }

    fn serialize_unit(self) -> Result<Self::Ok, D> {
        Ok(Value::simple_record("tuple", vec![]).wrap())
    }

    fn serialize_unit_struct(self, name: &'static str) -> Result<Self::Ok, D> {
        Ok(Value::simple_record(name, vec![]).wrap())
    }

    fn serialize_unit_variant(self,
                              _name: &'static str,
                              _variant: u32,
                              variant_name: &'static str) ->
        Result<Self::Ok, D>
    {
        Ok(Value::simple_record(variant_name, vec![]).wrap())
    }

    fn serialize_newtype_struct<T: ?Sized>(self, name: &'static str, value: &T) ->
        Result<Self::Ok, D> where T: Serialize
    {
        if name == crate::value::value::MAGIC {
            let v: crate::value::PlainValue<D> = to_value(value)?;
            let mut buf: &[u8] = v.value().as_bytestring().ok_or(Error::InternalMagicError)?;
            crate::value::Decoder::new(&mut buf, None).next().or(Err(Error::InternalMagicError))
        } else {
            // TODO: This is apparently discouraged, and we should apparently just serialize `value`?
            Ok(Value::simple_record(name, vec![to_value(value)?]).wrap())
        }
    }

    fn serialize_newtype_variant<T: ?Sized>(self,
                                            _name: &'static str,
                                            _variant: u32,
                                            variant_name: &'static str,
                                            value: &T) ->
        Result<Self::Ok, D> where T: Serialize
    {
        Ok(Value::simple_record(variant_name, vec![to_value(value)?]).wrap())
    }

    fn serialize_seq(self, count: Option<usize>) -> Result<Self::SerializeSeq, D> {
        let vec = match count { Some(n) => Vec::with_capacity(n), None => Vec::new() };
        Ok(SerializeSequence{ vec, phantom: PhantomData })
    }

    fn serialize_tuple(self, count: usize) -> Result<Self::SerializeTuple, D> {
        Ok(SerializeRecord{ name: "tuple", vec: Vec::with_capacity(count), phantom: PhantomData })
    }

    fn serialize_tuple_struct(self, name: &'static str, count: usize) ->
        Result<Self::SerializeTupleStruct, D>
    {
        Ok(SerializeRecord{ name, vec: Vec::with_capacity(count), phantom: PhantomData })
    }

    fn serialize_tuple_variant(self,
                               _name: &'static str,
                               _variant: u32,
                               variant_name: &'static str,
                               count: usize) ->
        Result<Self::SerializeTupleVariant, D>
    {
        Ok(SerializeRecord{ name: variant_name, vec: Vec::with_capacity(count), phantom: PhantomData })
    }

    fn serialize_map(self, _count: Option<usize>) -> Result<Self::SerializeMap, D> {
        Ok(SerializeDictionary{ next_key: None, items: Map::new(), phantom: PhantomData })
    }

    fn serialize_struct(self, name: &'static str, count: usize) -> Result<Self::SerializeStruct, D> {
        Ok(SerializeRecord{ name, vec: Vec::with_capacity(count), phantom: PhantomData })
    }

    fn serialize_struct_variant(self,
                                _name: &'static str,
                                _variant: u32,
                                variant_name: &'static str,
                                count: usize) ->
        Result<Self::SerializeStructVariant, D>
    {
        Ok(SerializeRecord{ name: variant_name, vec: Vec::with_capacity(count), phantom: PhantomData })
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeMap for SerializeDictionary<N, D> {
    type Ok = N;
    type Error = Error<D>;

    fn serialize_key<T: ?Sized>(&mut self, key: &T) -> Result<(), D> where T: Serialize {
        self.next_key = Some(to_value(key)?);
        Ok(())
    }

    fn serialize_value<T: ?Sized>(&mut self, value: &T) -> Result<(), D> where T: Serialize {
        let key = self.next_key.take().unwrap();
        self.items.insert(key, to_value(value)?);
        Ok(())
    }

    fn end(self) -> Result<Self::Ok, D> {
        Ok(Value::from(self.items).wrap())
    }
}

impl<N: NestedValue<D>, D: Domain> SerializeRecord<N, D> {
    fn push(&mut self, value: N) -> Result<(), D> {
        self.vec.push(value);
        Ok(())
    }

    fn finish(self) -> Result<N, D> {
        Ok(Value::simple_record(self.name, self.vec).wrap())
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeStruct for SerializeRecord<N, D> {
    type Ok = N;
    type Error = Error<D>;

    fn serialize_field<T: ?Sized>(&mut self, _name: &'static str, value: &T) -> Result<(), D>
    where T: Serialize
    {
        self.push(to_value(value)?)
    }

    fn end(self) -> Result<Self::Ok, D> {
        self.finish()
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeStructVariant for SerializeRecord<N, D> {
    type Ok = N;
    type Error = Error<D>;

    fn serialize_field<T: ?Sized>(&mut self, _name: &'static str, value: &T) -> Result<(), D>
    where T: Serialize
    {
        self.push(to_value(value)?)
    }

    fn end(self) -> Result<Self::Ok, D> {
        self.finish()
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeTuple for SerializeRecord<N, D> {
    type Ok = N;
    type Error = Error<D>;

    fn serialize_element<T: ?Sized>(&mut self, value: &T) -> Result<(), D> where T: Serialize {
        self.push(to_value(value)?)
    }

    fn end(self) -> Result<Self::Ok, D> {
        self.finish()
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeTupleStruct for SerializeRecord<N, D> {
    type Ok = N;
    type Error = Error<D>;

    fn serialize_field<T: ?Sized>(&mut self, value: &T) -> Result<(), D> where T: Serialize {
        self.push(to_value(value)?)
    }

    fn end(self) -> Result<Self::Ok, D> {
        self.finish()
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeTupleVariant for SerializeRecord<N, D> {
    type Ok = N;
    type Error = Error<D>;

    fn serialize_field<T: ?Sized>(&mut self, value: &T) -> Result<(), D> where T: Serialize {
        self.push(to_value(value)?)
    }

    fn end(self) -> Result<Self::Ok, D> {
        self.finish()
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeSeq for SerializeSequence<N, D> {
    type Ok = N;
    type Error = Error<D>;

    fn serialize_element<T: ?Sized>(&mut self, value: &T) -> Result<(), D> where T: Serialize {
        self.vec.push(to_value(value)?);
        Ok(())
    }

    fn end(self) -> Result<Self::Ok, D> {
        Ok(Value::from(self.vec).wrap())
    }
}

pub fn to_value<T, N, D: Domain>(value: T) -> Result<N, D> where T: Serialize, N: NestedValue<D> {
    value.serialize(Serializer{ phantom: PhantomData })
}
